////////////////// INCLUDES //////////////////

#include <config.h>

#include <conio.h>
#include <unistd.h>
#include <tm.h>

#include <dsensor.h>
#include <dmotor.h>
#include <dsound.h>

typedef unsigned char uchar;
typedef signed char schar;

#define TBS_MAXSIZE 256
// Includes the problem definition
#include "problemdefs/problemdef.h"

#include "scheduler_api.h"

////////////////// FORWARD DECLARATIONS (FUNCTION PROTOTYPES) //////////////////

void check_deadlines();

////////////////// CONSTANTS AND DEFINITIONS //////////////////


#define MOTOR_A 0
#define MOTOR_B 1
#define MOTOR_C 2

#define INFINITY (1<<15)

#define FALSE 0
#define TRUE 1

// Parameters for the object detector (checks rotation sensor ITERS times, with DELAY ms between, and detects deflection of TOL or greater)
#define OBJECT_CHECK_ITERS 15
#define OBJECT_CHECK_TOL 2
#define OBJECT_CHECK_DELAY 80

// Output port assignments
#define TRAIN_ENGINE	MOTOR_A
#define STATION_ARM	MOTOR_B
#define BIN_DUMP	MOTOR_C

// Input port assignments (XXX = sensor address, XXX_V = sensor value)
#define STATION_SENSOR		SENSOR_2
#define STATION_SENSOR_V	TOUCH_2
#define OBJECT_SENSOR   	SENSOR_3
#define OBJECT_SENSOR_V 	ROTATION_3

// Notes
note_t fail_notes[] = {{PITCH_E1, 4}, {PITCH_A0, 8}, {PITCH_END, 0}};
note_t accept_notes[] = {{PITCH_A3, 2}, {PITCH_C4, 2}, {PITCH_E4, 2}, {PITCH_END, 0}};
note_t besteffort_notes[] = {{PITCH_E4, 2}, {PITCH_C4, 2}, {PITCH_A3, 2}, {PITCH_END, 0}};
note_t has_obj_notes[] = {{PITCH_A5, 1}, {PITCH_PAUSE, 1}, {PITCH_A5, 1}, {PITCH_END, 0}};

// WCET calculation
const time_t WCET = ((NUM_STATIONS / 2) * UNIT_TRANSIT_TIME + CHECK_TIME);

////////////////// STATE VARIABLES //////////////////

// System clock at program start
time_t start_time;

uchar cur_pos; // Current station
time_t sttn_release[NUM_STATIONS]; // Release time of each station check task (index 0 ignored)
time_t sttn_deadline[NUM_STATIONS]; // Absolute deadline for each station check task (index 0 ignored)

time_t tbs_deadline; // TBS deadline variable
uchar has_tbs_task;  // Whether TBS has a pending task

time_t deliver_deadline; // Earliest delivery deadline
uchar has_object;        // Whether there is an object for guaranteed delivery

////////////////// UTILITY FUNCTIONS //////////////////

int abs(int x) {
  if (x < 0) return -x;
  return x;
}

void time_init() {
  start_time = get_system_up_time();
}

time_t get_current_time() {
  return get_system_up_time() - start_time;
}

////////////////// STATION/POSITION INQUIRY FUNCTIONS //////////////////

// Return the distance around the track to the specified station. Positive return
// value means forward travel, negative means reverse. Returns the direction with
// the lowest absolute distance (ties go to forward/positive)
schar get_delta_to_station(uchar dest) {
	schar fwddist, revdist;
	fwddist = cur_pos <= dest ? (schar)dest - (schar)cur_pos: (schar)dest - (schar)cur_pos + (schar)NUM_STATIONS;
	revdist = dest <= cur_pos ? (schar)dest - (schar)cur_pos : (schar)dest - (schar)cur_pos - (schar)NUM_STATIONS;

	return fwddist <= -revdist ? fwddist : revdist;
}

schar get_delta_between_stations(uchar start, uchar dest) {
	schar fwddist, revdist;
	fwddist = start <= dest ? (schar)dest - (schar)start: (schar)dest - (schar)start + (schar)NUM_STATIONS;
	revdist = dest <= start ? (schar)dest - (schar)start : (schar)dest - (schar)start - (schar)NUM_STATIONS;

	return fwddist <= -revdist ? fwddist : revdist;
}

// Adds a delta to a given pos. Delta must be in [-(NUM_STATIONS - 1), NUM_STATIONS - 1]
inline uchar add_to_pos(uchar pos, schar delta) {
	schar newpos = (schar)pos + delta;
	if (newpos < 0) newpos += NUM_STATIONS;
	else if (newpos >= NUM_STATIONS) newpos -= NUM_STATIONS;
	return (uchar)newpos;
}

////////////////// MECHANISM FUNCTIONS //////////////////

//
// Sensor functions and callbacks
//

wakeup_t station_sensor_wakeup(wakeup_t data) {
	return data ? STATION_SENSOR_V : !STATION_SENSOR_V; 
}

inline void wait_station_sensor_on() { wait_event(&station_sensor_wakeup, 1); }
inline void wait_station_sensor_off() { wait_event(&station_sensor_wakeup, 0); }

//
// Basic motor control
//

void motor_speed(uchar motor, uchar speed) {
	if (motor == MOTOR_A) motor_a_speed(speed);
	else if (motor == MOTOR_B) motor_b_speed(speed);
	else if (motor == MOTOR_C) motor_c_speed(speed);
}

void motor_dir(uchar motor, MotorDirection dir) {
	if (motor == MOTOR_A) motor_a_dir(dir);
	else if (motor == MOTOR_B) motor_b_dir(dir);
	else if (motor == MOTOR_C) motor_c_dir(dir);
}

void motor_pulse(uchar motor_id, uchar speed, int direction, int duration) {
	int i;
	uchar rawspeed = (uchar)((int)speed*MAX_SPEED/100);

	for (i = 0; i < duration; i++) {
		motor_speed(motor_id, rawspeed);
		motor_dir(motor_id, direction);
		msleep(100);
		motor_speed(motor_id, brake);
		msleep(50);
	}
	motor_dir(motor_id, brake);
	msleep(200);
	motor_dir(motor_id, off);
}

//
// Mid-level movement routines
//

char wait_for_objects(time_t until) {
  char i, has_obj = FALSE;
  int val;
  ds_rotation_set(&OBJECT_SENSOR, 0);
  for (i = 0; i < OBJECT_CHECK_ITERS || get_current_time() + OBJECT_CHECK_DELAY < until; i++) { // Don't stop if we have more than OBJECT_CHECK_DELAY until the stoptime
    val = OBJECT_SENSOR_V;
    if (val > OBJECT_CHECK_TOL || val < -OBJECT_CHECK_TOL) {
      if (!has_obj) dsound_play(has_obj_notes);
      has_obj = TRUE;
    }
    msleep(OBJECT_CHECK_DELAY);
  }

  return has_obj;
}

char tip_input_bin(time_t until) {
	char ret;
	motor_pulse(STATION_ARM, 35, fwd, 2);
    motor_speed(STATION_ARM, 30); // 30/255
    motor_dir(STATION_ARM, fwd);
	ret = wait_for_objects(until > 700 ? until - 700 : 0); // -700 = time to retract arm
	motor_pulse(STATION_ARM, 20, rev, 3);
	return ret;
}

void dump(){
	motor_pulse(BIN_DUMP, 40, fwd, 8);
	msleep(1000);
	motor_pulse(BIN_DUMP, 20, rev, 3);
}

//
// High-level action functions
//

// Moves the train by the given delta (in stations)
void move_delta(schar delta) {
	MotorDirection dir = delta < 0 ? rev : fwd;

	if (delta == 0) return;

	// Move train forward at full power to get up to speed
	motor_dir(TRAIN_ENGINE, dir);
	motor_speed(TRAIN_ENGINE, MAX_SPEED);

	do {
		// Wait to ensure we are off any station indicator strips, then reduce
		// power so we don't go off the rails of the crazy train
		msleep(400);
		motor_speed(TRAIN_ENGINE, MAX_SPEED*27/40);

		// Wait until we hit a station indicator strip
		wait_station_sensor_on();

		// Update position
		cur_pos = add_to_pos(cur_pos, dir == fwd ? 1 : -1);
		delta += dir == fwd ? -1 : 1;
	} while (delta != 0);

	// Apply the brakes
	motor_dir(TRAIN_ENGINE, brake);
	msleep(300);

	motor_speed(TRAIN_ENGINE, MAX_SPEED*13/20);

	// Align with the station. Goal: position the train at the forward end of the strip.

	// If we were going in reverse and overshot the strip, move forward to the beginning of the strip.
	if (dir == rev && !STATION_SENSOR_V) {
		motor_dir(TRAIN_ENGINE, fwd);
		msleep(150);
		wait_station_sensor_on();
		msleep(75);
		motor_dir(TRAIN_ENGINE, brake);
	}
	
	// Now we are either on the strip or past the forward end (if we were going forward).
	// If the former, move past the forward end of the strip.
	if (STATION_SENSOR_V) {
		motor_dir(TRAIN_ENGINE, fwd);
		msleep(150);
		wait_station_sensor_off();
		msleep(75);
		motor_dir(TRAIN_ENGINE, brake);
	}

	// Now we are past the forward end of the strip. Just move backward until we hit it, and we are done.
	motor_dir(TRAIN_ENGINE, rev);
	msleep(75);
	wait_station_sensor_on();
	motor_dir(TRAIN_ENGINE, brake);
	msleep(75);
}

// Moves to the specified station and performs the correct action
// depending on station type (input or output bin)
uchar service_station(uchar sttn) {
  // Move to the specified station
  char new_obj;
  schar delta = get_delta_to_station(sttn);
  move_delta(delta);

  // Perform the correct action at that station (dump or check), then check whether
  // any deadlines have been missed
  if (sttn == 0) {
    dump();
    check_deadlines();

    has_object = FALSE;

    return FALSE;
  } else {
    // If we arrive early, wait until just before we need to start
    // the checking procedure to be finished right as the task releases
    time_t now = get_current_time();
    if (sttn_release[sttn] > now + CHECK_TIME)
      msleep(sttn_release[sttn] - now - CHECK_TIME);

    // Now do the bin check, and be absolutely sure to finish on or
    // after the release time!
    new_obj = tip_input_bin(sttn_release[sttn]);
    check_deadlines();

    // Update to new release/deadline
    sttn_release[sttn] += sttn_period[sttn];
    sttn_deadline[sttn] += sttn_period[sttn];

    return new_obj;
  }
}

////////////////// SCHEDULER FUNCTIONS //////////////////

//
// Deadline checking
//

char deadline_missed() {
  uchar i;
  time_t time = get_current_time();

  // Check periodic task deadlines
  for (i = 1; i < NUM_STATIONS; i++) {
    if (time >= sttn_deadline[i]) {
      cputw(i);
      return TRUE;
    }
  }

  // Check TBS deadline
  if (has_tbs_task && time > tbs_deadline) {
    cputs("tbs");
    return TRUE;
  }

  // Check guaranteed object delivery time deadlines (this is actually
  // redundant with TBS, as TBS is used to make this guarantee, but we
  // include it for debugging/extra provenance of correctness)
  if (has_object && time > deliver_deadline) {
    cputs("out");
    return TRUE;
  }

  return FALSE;
}

void check_deadlines() {
  if (deadline_missed()) {
    dsound_play(fail_notes);
    msleep(5000);
    exit(1);
  }
}

//
// Scheduler init
//

void scheduler_init() {
  uchar i;

  cur_pos = 0;
  for (i = 1; i < NUM_STATIONS; i++) {
    sttn_release[i] = 0;
    sttn_deadline[i] = sttn_period[i];
  }

  tbs_deadline = 0;
  has_tbs_task = FALSE;
  deliver_deadline = INFINITY;
  has_object = FALSE;
}

//
// Scheduler API
//

#include "train_edf_sched.c"

////////////////// MAIN //////////////////

int main(int argc, char *argv[]) {
  // Activate sensors as necessary
  ds_active(&OBJECT_SENSOR);
  ds_rotation_on(&OBJECT_SENSOR);

  // Initialize
  time_init();
  scheduler_init();

  // Infinite scheduling loop!
  while(!shutdown_requested()) {
    // Ask scheduler for next move, then do it
    uchar sttn = scheduler_get_next_move();
    uchar newobj = service_station(sttn);

    // If we just picked up an object, inform the scheduler
    // Else, if we just dropped off our load, inform the scheduler
    if (newobj) {
      uchar guarantee = scheduler_process_new_object(sttn);

      // Play accept or reject/best-effort sound depending on scheduler's response
      if (guarantee) {
        // Lock in object delivery guarantee for deadline checking
        time_t new_dd = get_current_time() + sttn_obj_deadline[sttn];
        if (!has_object || new_dd < deliver_deadline) deliver_deadline = new_dd;
        has_object = TRUE;

        dsound_play(accept_notes);
      } else {
        // No guarantee made; just play a sound
        dsound_play(besteffort_notes);
      }
    } else if (cur_pos == 0) {
      scheduler_process_object_dump();
    }
  }

  return 0;
}

