typedef struct {
  uchar sttn;
  time_t deadline;
} edf_task_t;

edf_task_t get_earliest_deadline_task() {
  time_t ed = 0;            // Earliest deadline
  uchar eds = NUM_STATIONS; // Earliest deadline station (start with invalid value)
  time_t now = get_current_time();

  uchar sttn;

  // If we have an object for delivery, the TBS has a task, so start with that
  if (has_tbs_task) {
    ed = tbs_deadline;
    eds = 0;
  }

  for (sttn = 1; sttn < NUM_STATIONS; sttn++) {
#ifdef ADV_EDF
    if (sttn_release[sttn] > now + WCET) continue;
#else
    if (sttn_release[sttn] > now) continue;
#endif

    // If we have the earliest deadline (or no other task has been considered yet), update
    if (eds == NUM_STATIONS || sttn_deadline[sttn] < ed) {
      ed = sttn_deadline[sttn];
      eds = sttn;
    }
  }

  return (edf_task_t){eds, ed};
}

#ifdef ADV_TBS
uchar get_farthest_station_before_deadline(time_t max_deadline) {
  uchar fsd = 0; // Farthest station distance
  uchar sttn, dist;
  time_t now = get_current_time();

  if (has_tbs_task) {
    fsd = abs(get_delta_to_station(0));
  }

  for (sttn = 1; sttn < NUM_STATIONS; sttn++) {
#ifdef ADV_EDF
    if (sttn_release[sttn] > now + WCET) continue;
#else
    if (sttn_release[sttn] > now) continue;
#endif

    if (sttn_deadline[sttn] > max_deadline) continue;

    dist = abs(get_delta_to_station(sttn));
    if (dist > fsd) fsd = dist;
  }

  return fsd;
}
#endif ADV_TBS

// EDF implementation
// Note: this function MAY return a task/station whose release time has not
// yet occured. In this case, the system will synchronize the train such that
// it will not complete its check before the release
uchar scheduler_get_next_move() {
#ifndef ADV_EDF
  time_t now = get_current_time();
#endif
  time_t er = sttn_release[1]; // Earliest release
  uchar ers = 1;               // Earliest release station

  uchar sttn;  

  // If some task is currently released, go with the one with the earliest deadline
  edf_task_t task = get_earliest_deadline_task();
  if (task.sttn != NUM_STATIONS)
    return task.sttn;

  // Else, find the periodic task with the minimum release time and execute that
  // one once it is released
  for (sttn = 2; sttn < NUM_STATIONS; sttn++) {
    if (sttn_release[sttn] < er) {
      er = sttn_release[sttn];
      ers = sttn;
    }
  }

#ifndef ADV_EDF
  // Sleep until the earliest release, then return to go there
  // (Since we are not allowed to do pre-release moves without ADV_EDF)
  msleep(er - now);
#endif
  return ers;
}

uchar scheduler_process_new_object(uchar sttn) {
  time_t now = get_current_time();
  time_t new_obj_deadline = now + sttn_obj_deadline[sttn];

#ifdef ADV_TBS
  if (!has_tbs_task) {
    time_t tent_tbs_deadline;
    time_t tent_exec_time;
    uchar last_dist, max_dist;

    if (tbs_deadline < now)
       tbs_deadline = now;    // For simplicity of calculation below

     // Iteratively compute the maximum distance the delivery task would have
     // to cover based on all periodic station tasks which have deadlines
     // earlier than the TBS deadline (which is iteratively updated)
     last_dist = 0;
     max_dist = abs(get_delta_between_stations(sttn, 0));
     do {
       last_dist = max_dist;

       // Compute new tentative WCET and TBS deadline
       tent_exec_time = (time_t)max_dist * UNIT_TRANSIT_TIME + CHECK_TIME;
       tent_tbs_deadline = tbs_deadline + tent_exec_time * TBS_MAXSIZE / TBS_SIZE; // exectime * 1/U_s

       // Get farthest station which has a deadline before this
       max_dist = get_farthest_station_before_deadline(tent_tbs_deadline);
     } while (max_dist > last_dist);

     // Use the final tentative TBS deadline as the actual
     tbs_deadline = tent_tbs_deadline;
     has_tbs_task = TRUE;
  }
  return tbs_deadline <= new_obj_deadline;
#else
  // If this is the first object, add a TBS task for it
  // Else, just check whether the new object fits the already-established TBS task deadline
  if (!has_tbs_task) {
    if (tbs_deadline < now)
       tbs_deadline = now;    // For simplicity of calculation below

    tbs_deadline += WCET * TBS_MAXSIZE / TBS_SIZE; // WCET * 1/U_s
    has_tbs_task = TRUE;
  }
  return tbs_deadline <= new_obj_deadline;
#endif
}

void scheduler_process_object_dump() {
  has_tbs_task = FALSE;
}

